import re

from markdown.util import etree
from markdown.inlinepatterns import Pattern
from markdown.preprocessors import Preprocessor
from markdown.extensions import Extension


class Todo(Extension):

    """ == TODO, FIXME, WARNING, CAUTION marks == """

    class Prep(Preprocessor):

        """
        Markdown preprocessor that matches all TODO and FIXME strings occurring at the beginning\
        of the line or after the inline comment delimiter and highlights them in the documentation.
        """

        matched_strings = ["TODO", "FIXME", "WARNING", "CAUTION"]
        regex = re.compile("^\s*(" + "|".join(matched_strings) + ":?)(.*)", flags=re.I)

        def template(self, match):
            """ Intended markup for TODO strings. The type of the string is used as a class. """
            return "<span class={0:s}><strong>{1:s}</strong>{2:s}</span>"\
                .format(match.group(1).lower(), match.group(1), match.group(2))

        def run(self, lines):
            """ String matching is case insensitive """
            return [self.regex.sub(self.template, line) for line in lines]

    def extendMarkdown(self, md, md_globals):
        md.preprocessors.add('todo', Todo.Prep(md), '_end')


class LinesConnector(Extension):

    """ == Lines connector extension == """

    def __init__(self, regex=r"(\S)\s*\\\s*\n\s*(\S)", sub=r"\1 \2", *args, **kwargs):

        regex = re.compile(regex, flags=re.M)

        class Prep(Preprocessor):

            def run(self, lines):
                """ Method ensures that there is exactly one space between 2 joined strings. """
                return regex.sub(sub, "\n".join(lines)).split("\n")

        self.Prep = Prep()

        super(LinesConnector, self).__init__(*args, **kwargs)

    def extendMarkdown(self, md, md_globals):
        md.preprocessors.add('lines-connector', self.Prep, '_end')


class SaneDefList(Extension):

    """ == Better definition lists == """

    class Prep(Preprocessor):

        """
        Markdown preprocessor that prepares natural-style definition lists for native Markdown \
        extension `def_list`. It allows to write more compact and readable class field definitions.
        """

        def run(self, lines):
            """
            Searches for a line starting with a literal followed by a colon and multiple spaces:

                some arbitrary parameter name:  and its definition separated by `:\s\s+`
                `parameter name` might contain\
                everything except a \
                [colon](//en.wikipedia.org/wiki/Colon_(punctuation))\
                and be multiline:               it still works

            And turns it into:

            some arbitrary parameter name:  and its definition separated by `:\s\s+`
            `parameter name` might contain\
            everything except a \
            [colon](//en.wikipedia.org/wiki/Colon_(punctuation))\
            and be multiline:               it still works
            """
            new_lines = []
            for line in lines:
                match = re.match(r'^(\s*)([^:]+):\s{2,}(.+)', line)
                if match:
                    new_lines.append(match.group(1) + match.group(2))
                    new_lines.append(match.group(1) + ':   ' + match.group(3))
                    new_lines.append('')
                else:
                    new_lines.append(line)

            return new_lines

    def extendMarkdown(self, md, md_globals):
        md.preprocessors.add('sane-def-list', SaneDefList.Prep(md), '_end')


class Pydoc(Extension):

    """ == Docblocks meta marks processor == """

    class Prep(Preprocessor):

        """ Preprocessor used to parse PyDoc-style comments like `:param name:` and format them. """

        def run(self, lines):
            """
            :param lines: Documentation lines
            :return: Lines of text with parsed PyDoc comments
            """
            new_lines = []
            for text in lines:
                # Regex that matches `@param name`
                text = re.compile(
                    r'^(\s?)@(\w+)\s+(["\'\`].+["\'\`]|\S+)\s*', re.M
                ).sub(
                    r'\1<span class="pydoc pydoc-\2"><span>\2</span> <code>\3</code></span> ',
                    text)
                # Regex that matches `@var`
                text = re.compile(
                    r'^(\s?)@(\w+)', re.M
                ).sub(
                    r'\1<span class="pydoc pydoc-\2"><span>\2</span></span>',
                    text)
                # Regex that matches `:param name:`
                text = re.compile(
                    r'^(\s?):([^: ]+) +([^:]+):', re.M
                ).sub(
                    r'\1<span class="pydoc pydoc-\2"><span>\2</span> <code>\3</code></span>',
                    text)
                # Regex that matches single-word comments: `:return:`
                text = re.compile(
                    r'^(\s?):([^: ]+):', re.M
                ).sub(
                    r'\1<span class="pydoc pydoc-\2"><span>\2</span></span>',
                    text)

                new_lines.append(text)
            return new_lines

    def extendMarkdown(self, md, md_globals):
        md.preprocessors.add('pydoc', Pydoc.Prep(md), '_end')


class AutoLinkExtension(Extension):

    """ == Autolink extension ==
        There's already an inline pattern called autolink which handles\
        <http://www.google.com> type links. """

    EXTRA_AUTOLINK_RE = r'(?<!"|>)((https?://|www)[-\w./#?%=&]+)'

    class pattern(Pattern):

        def handleMatch(self, m):
            el = etree.Element('a')
            if m.group(2).startswith('http'):
                href = m.group(2)
            else:
                href = 'http://%s' % m.group(2)
            el.set('href', href)
            el.text = m.group(2)
            return el

    def extendMarkdown(self, md, md_globals):
        md.inlinePatterns.add(
            'extra_autolink',
            AutoLinkExtension.pattern(self.EXTRA_AUTOLINK_RE, self),
            '>autolink'
        )



"""
Read Pycco static resources into module variables
"""
import os

with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "pycco.html"), 'rb') as f:
    html = f.read()

static_files = [
    ('pyccoon.svg', 'pyccoon.svg'),
    ('pyccoon_icon.svg', 'pyccoon_icon.svg'),
    ('pycco.css', 'pyccoon.css')
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
![Pyccoon](pyccoon.svg)

"**Pyccoon**" is a side-to-side documentation generator.
"""

import optparse
import os
import shutil
import pystache
import re
import sys
import json
from io import open
from datetime import datetime
from collections import defaultdict

# This module contains all of our static resources.
from . import resources, __version__, __author__
from .languages import get_language, Language

from .utils import shift, ensure_directory


# == Main documentation generation class ==

class Pyccoon:

    add_lineno = True

    # The start of each Pygments highlight block.
    highlight_start = "<div class=\"highlight\"><pre>"

    # The end of each Pygments highlight block.
    highlight_end = "</pre></div>"

    config = defaultdict(lambda: [], {
        "skip_files": [".+\\.pyc", "__pycache__", "\\.travis.yml", "\\.git", "\\.DS_Store"]
    })

    config_file = '.pyccoon'
    watch = False
    verbosity = -1

    def __init__(self, opts, process=True):
        """
        == Pyccoon initialization ==
        :param opts: `dict` of parameters.
        :param process: Whether to generate documentation immediately

        Available parameters:

          * `sourcedir` - project source directory
          * `outdir` - output directory
          * `config_file` - pyccoon project settings
          * `watch` - whether to regenerate the docs automatically
        """

        for key, value in opts.items():
            setattr(self, key, value)

        self.log("Pyccoon {0}".format(__version__))
        self.log("-------------")

        self.init_config()

        self.verbosity = self.config['verbosity'] or 1 if self.verbosity == -1 else self.verbosity

        if not self.outdir:
            raise TypeError("Missing the required 'outdir' argument.")
        if not self.sourcedir:
            raise TypeError("Missing the required 'sourcedir' argument.")

        self.sourcedir = os.path.abspath(self.sourcedir)
        self.log("Source folder: " + self.sourcedir)
        self.outdir = os.path.abspath(self.outdir)
        self.log("Output folder: " + self.outdir)

        self.collect_sources()

        # Create the template that we will use to generate the Pyccoon HTML page.
        self.page_template = self.template(resources.html)

        if process:
            self.process()

        # If the -w / --watch option was present, monitor the source directories \
        # for changes and re-generate documentation for source files whenever they \
        # are modified.
        if self.watch:
            try:
                import watchdog.events
                import watchdog.observers
            except ImportError:
                sys.exit('The `watch` option requires the watchdog package.')

            from .utils import monitor
            monitor(path=self.sourcedir, func=lambda: self.process())

    def log(self, message):
        if self.verbosity:
            print(message)

    def init_config(self):
        """ Try to get `.pyccoon` config file or use the default values """
        config_file = os.path.abspath(self.config_file)
        if os.path.exists(config_file):
            self.log('Using config {0:s}'.format(config_file))
            with open(config_file, 'rb') as f:
                self.config.update(json.loads(f.read().decode('utf8')))

        self.config['skip_files'] = [re.compile(p) for p in self.config['skip_files']]
        self.config['copy_files'] = [re.compile(p) for p in self.config['copy_files']]

        self.project_name = self.config['project_name'] \
            or (os.path.split(self.sourcedir)[1] + " documentation")

    def collect_sources(self):
        """ Collect names of all files to be copied or processed """
        self.sources = {}
        for dirpath, dirnames, files in os.walk(self.sourcedir):
            if any([regex.search(dirpath) for regex in self.config['skip_files']]):
                continue
            for name in files:
                if not any([regex.search(name) for regex in self.config['skip_files']]) \
                        and name not in dirnames:
                    source = os.path.relpath(os.path.join(dirpath, name), self.sourcedir)
                    process = True
                    if any([regex.search(name) for regex in self.config['copy_files']]):
                        process = False

                    self.sources[source] = (self.destination(source), process)

    def process(self, sources=None, language=None):
        """
        == Source files processing ==

        :param sources: `list` of source files to process
        :param language: Force programming language
        """

        self.log('\n' + '-'*80)
        self.log("[{0}] Generating documentation for {1}".format(datetime.now(), self.project_name))
        self.log('-'*80 + '\n')

        if sources:
            sources = dict([(k, v) for (k, v) in self.sources.items() if k in sources])
        else:
            sources = self.sources

        ensure_directory(self.outdir)

        for file, dest in resources.static_files:
            filepath = os.path.join(os.path.split(resources.__file__)[0], file)
            destpath = os.path.join(self.outdir, dest)
            self.sources[filepath] = (destpath, False)
            shutil.copyfile(
                filepath,
                destpath
            )

        # Proceed to generating the documentation.
        for source, (dest, process) in sorted(sources.items(), key=lambda x: x[0]):

            try:
                if process:
                    with open(os.path.join(self.sourcedir, source), "rb") as sourcefile:
                        code = sourcefile.read().decode('utf8')

                    self.language = get_language(source, code, language=language)
                    if not self.language:
                        process = False
                        self.sources[source] = (dest, process)

                    try:
                        ensure_directory(os.path.split(dest)[0])
                    except OSError:
                        pass

                if process:
                    if os.path.exists(os.path.join(self.sourcedir, source)):
                        with open(dest, "wb") as f:
                            f.write(self.generate_documentation(source, code,
                                                                language=self.language)
                                    .encode('utf8'))

                        self.log("\tProcessed:\t{0:s} -> {1:s}"
                                 .format(source, os.path.relpath(dest, self.outdir)))
                    else:
                        self.log("File does not exist: {0:s}".format(source))

                else:
                    ensure_directory(os.path.split(dest)[0])
                    shutil.copyfile(os.path.join(self.sourcedir, source), dest)
                    self.log("\tCopied:   \t{0:s}".format(source))
            except Exception as e:
                self.log("Error while processing file {0:s}: {1}".format(source, e))

        # Ensure there is always an index file in the output folder
        for _, (dest, _) in list(self.sources.items()):
            folder = os.path.relpath(os.path.split(dest)[0], self.outdir).lstrip('./')
            index = os.path.join(folder, "index.html")

            if not any([os.path.join(self.outdir, index) == dest
                        for _, (dest, _) in self.sources.items()]):
                source = os.path.join(folder, 'index.html')
                self.sources[source] = (os.path.join(self.outdir, index), False)

                with open(os.path.join(self.outdir, index), 'w', encoding='utf8') as f:
                    self.language = Language()
                    f.write(self.generate_html(source, []))
                    self.log("\tGenerated:\t{0:s}".format(source))

    def template(self, source):
        return lambda context: pystache.render(source, context)

    def generate_documentation(self, source, code, language=None):
        """
        == Generating documentation ==
        Generate the documentation for a source file by reading it in, splitting it\
        up into comment/code sections, highlighting them for the appropriate\
        language, and merging them into an HTML template.
        """

        self.sections = language.parse(code, add_lineno=self.add_lineno)
        self.highlight(source, self.sections, language)
        return self.generate_html(source, self.sections)

    def highlight(self, source, sections, language):
        """
        === Highlighting the source code ===

        Highlights a single chunk of code using the **Pygments** module, and runs\
        the text of its corresponding comment through **Markdown**.

        We process the entire file in a single call to Pygments by inserting little\
        marker comments between each section and then splitting the result string\
        wherever our markers occur.
        """
        output = language.highlight(
            language.divider_text.join(section["code_text"].rstrip() for section in sections)
        )

        output = output.replace(self.highlight_start, "").replace(self.highlight_end, "")
        fragments = re.split(language.divider_html, output)
        for i, section in enumerate(sections):
            section["code_html"] = shift(fragments, "")
            if section["code_html"]:
                section["code_html"] = \
                    self.highlight_start + section["code_html"] + self.highlight_end
            docs_text = section["docs_text"]
            section["docs_html"] = language.markdown(
                self.preprocess(docs_text, i, source=os.path.join(self.sourcedir, source))
            )
            section["num"] = i

    def preprocess(self, comment, section_nr, source):
        """
        === Preprocessing the comments ===

        Add cross-references before having the text processed by markdown.  It's\
        possible to reference another file, like this : `[[utils.py]]` which renders\
        [[utils.py]]. You can also reference a specific section of another file, like\
        this: `[[utils.py#ensure-directory]]` which renders as\
        [[utils.py#ensure-directory]]. Sections have to be manually\
        declared; they are written on a single line, and surrounded by equals signs:\
        `=== like this ===`
        """

        def slugify(name):
            """ Return URL-friendly section name representation """
            return "-".join(name.lower().strip().split(" "))

        def replace_crossref(match):
            name = match.group(1)
            if name:
                name = name.rstrip("|")
            path = match.group(2)

            if not name and not path:
                return

            # Check if the match contains an anchor
            anchor = None
            if '#' in path:
                path, anchor = path.split('#')

            if not name:
                name = os.path.basename(path)
                if anchor:
                    name = name + '#' + anchor

            anchor = '#' + anchor if anchor else ''

            if not path.startswith('.'):
                # Absolute reference
                path = os.path.relpath(
                    self.destination(path),
                    os.path.split(self.sources[os.path.relpath(source, self.sourcedir)][0])[0]
                )
            else:
                # Relative reference
                path = self.destination(
                    os.path.join(os.path.split(os.path.relpath(source, self.sourcedir))[0], path)
                )

            return "[{0:s}]({1:s}{2:s})".format(name, path, anchor)

        def replace_section_name(match):
            return (
                '\n{lvl} <a id="{id}" class="header-anchor" href="#{id}">{name}</a>'
            ).format(**{
                "lvl":  match.group(2).replace('=', '#'),
                "id":   slugify(match.group(3)),
                "name": match.group(3)
            })

        comment = re.compile(r'^\s*(#\s)?\s*([=]+)([^=]+)([=]+)\s*$', re.M)\
            .sub(replace_section_name, comment)
        comment = re.sub(r'\[\[([^\|\n]+\|)?(.+?)\]\]', replace_crossref, comment)

        return comment

    # == HTML Code generation ==

    def generate_html(self, source, sections):
        """
        Once all of the code is finished highlighting, we can generate the HTML file\
        and write out the documentation. Pass the completed sections into the\
        template found in `resources/pyccoon.html`.

        Pystache will attempt to recursively render context variables, so we must\
        replace any occurences of `{{`, which is valid in some languages, with a\
        "unique enough" identifier before rendering, and then post-process the\
        rendered template and change the identifier back to `{{`.
        """

        dest = self.destination(source)
        title = os.path.relpath(source, self.sourcedir)
        page_title = self.project_name + ": " + os.path.relpath(source, self.sourcedir).lstrip('./')
        csspath = os.path.relpath(os.path.join(self.outdir, "pyccoon.css"), os.path.split(dest)[0])

        breadcrumbs = self.generate_breadcrumbs(dest, title)
        children = self.generate_navigation(source)
        contents = self.generate_contents(sections)

        for section in sections:
            section['line_count'] = (section['code_text'].rstrip('\n') + '\n').count('\n')
            section['linenos'] = '\n'.join(str(section['line'] + i)
                                           for i in range(section['line_count']))

        rendered = self.page_template({
            "title":            page_title,
            "breadcrumbs":      breadcrumbs,
            "children":         children,
            "stylesheet":       csspath,
            "sections":         sections,
            "source":           source,
            "contents":         contents,
            "contents?":        bool(contents),
            "destination":      dest,
            "generation_time":  datetime.now(),
            "root_path":        os.path.relpath(".", os.path.split(source)[0]),
            "project_name":     self.project_name,
            "mathjax?":          self.config['mathjax']
        })

        return rendered.replace("__DOUBLE_OPEN_STACHE__", "{{")

    def generate_breadcrumbs(self, dest, title):
        """
        === Generating breadcrumbs ===
        Based on the source file path, generate linked breadcrumbs of the documentation.
        """
        breadcrumbs = []
        crumbpath = None

        dest_chunks = os.path.relpath(dest, self.outdir).split("/")
        source_chunks = title.split("/")
        dest_chunks.reverse()
        source_chunks.reverse()

        for i, crumb in enumerate(dest_chunks):

            crumbpath = os.path.join(crumbpath, "..") if crumbpath else crumb

            breadcrumbs.insert(0, {
                "title": source_chunks[i],
                "path": crumbpath if crumbpath.endswith('.html')
                else os.path.join(crumbpath, 'index.html')
            })
        breadcrumbs.insert(0, {
            "title": ".",
            "path": os.path.join(crumbpath, "../index.html")
        })

        return breadcrumbs

    def generate_navigation(self, source):
        """
        === Generating navigation ===
        For `index.html` files, generate a menu of folder contents.

        TODO: remove language dependency
        """
        index_names = ['__init__.py', 'index.html', 'index.php']
        if os.path.basename(source) not in index_names:
            return []

        children = []
        folder = os.path.split(os.path.join(self.sourcedir, source))[0]
        relfolder = os.path.relpath(folder, self.sourcedir)
        for filename in os.listdir(folder):
            if not any([regex.search(filename) for regex in self.config['skip_files']]):
                isdir = False
                filepath = None

                if os.path.isdir(os.path.join(folder, filename)):
                    isdir = True
                    filepath = os.path.join(filename, "index.html")
                else:
                    if filename in index_names:
                        filepath = "index.html"
                    elif os.path.join(relfolder, filename).lstrip("./") in self.sources:

                        filepath = filename
                        if self.sources[os.path.join(relfolder, filename).lstrip("./")][1]:
                            filepath = filepath + ".html"
                    else:
                        continue

                if filepath:
                    children.append({
                        "title": filename,
                        "path": filepath,
                        "isdir": isdir
                    })

        return sorted(children, key=lambda x: not x['isdir'])

    def generate_contents(self, sections):
        """
        === Generating page contents ===
        Gather the names of the documentation sections for "jump-to"-like navigation on the page.
        """
        contents = []

        for section in sections:
            section["code_html"] = section["code_html"].replace("{{", "__DOUBLE_OPEN_STACHE__")

            match = re.search(r'<h(\d)>(.+href=\"#(.+)\".+)</h(\d)>', section["docs_html"], re.M)

            if match:
                contents.append({
                    "url": "#{0}".format(match.group(3)),
                    "basename": re.sub(r'<[^<]+?>', '', match.group(2)),
                    "level": match.group(1)
                })
        return contents

    # == Utilities ==

    def destination(self, source, language=None):
        """
        Compute the destination HTML path for an input source file path. If the \
        source is `lib/example.py`, the HTML will be at `docs/lib/example.html`
        """

        dirname, filename = os.path.split(source)
        language = language or self.get_language(source)

        if language:
            name = language.transform_filename(filename)
        else:
            name = filename
        return os.path.normpath(os.path.join(self.outdir, os.path.join(dirname, name)))

    def get_language(self, source):
        """ Determine language of the file """
        language = None

        try:
            with open(os.path.join(self.sourcedir, source), "rb") as sourcefile:
                code = sourcefile.read().decode('utf8')

            language = get_language(source, code)
        except:
            pass

        return language


def main():
    """Hook spot for the console script."""

    parser = optparse.OptionParser(version='Pyccoon {0}'.format(__version__))
    parser.add_option('-s', '--source', action='store', type='string',
                      dest='sourcedir', default='.',
                      help='Source files directory (default: `%default`)')

    parser.add_option('-d', '--destination', action='store', type='string',
                      dest='outdir', default='docs',
                      help='Output directory (default: `%default`)')

    parser.add_option('-w', '--watch', action='store_true',
                      help='Watch original files and regenerate documentation on changes')

    parser.add_option('-c', '--config', action='store', dest='config_file',
                      default=os.path.join(os.getcwd(), '.pyccoon'), type='string',
                      help='Config file to use (default: `%default`)')

    parser.add_option('-v', '--verbosity', action='store', dest='verbosity',
                      default=-1, type='int',
                      help='Terminal output verbosity (0 to 1; default: %default)')

    opts, _ = parser.parse_args()
    opts = defaultdict(lambda: None, vars(opts))

    Pyccoon(opts)

# Run the script.
if __name__ == "__main__":
    main()



class Section(dict):

    """ Helper class that includes some frequently used routines """

    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

        if not self.get('line'):
            self['line'] = 1

    def has_code(self):
        """ Check if there is some code """
        return bool(self["code_text"].strip())

    def has_docs(self):
        """ Check if there are some docs """
        return bool(self["docs_text"].strip())

    def __missing__(self, key):
        """ Emulate `collections.defaultdict` behavior """
        return ''

    def copy(self):
        return Section(**dict((k, v) for (k, v) in self.items()))


# == Parsing strategy ==


class ParsingStrategy(list):

    """ Helper class that handles a list of methods and allows to insert item before or after \
        some specific item. """

    def __init__(self, *args):
        """ Initialize from an arbitrary sequence of arguments """
        for arg in args:
            self.append(arg)

    def index(self, name):
        """ Return the index of a method which name is `name` or raise an exception. """
        index = next((i for i, x in enumerate(self) if x.__name__ == name), None)
        if not index:
            raise Exception("Step '{0}' not found in the strategy: {1}".format(name, self))

        return index

    def insert_before(self, key, method):
        """ Insert a `method` before an item with a name `key` """
        self.insert(self.index(key), method)

    def insert_after(self, key, method):
        """ Insert a `method` after an item with a name `key` """
        self.insert(self.index(key) + 1, method)

    def delete(self, key):
        """ Remove the item with a name `key` """
        self.pop(self.index(key))


def iterate_sections(start=1, increment=1):
    """
    Helper decorator to iterate through the `sections` while altering them.

    :param start: Section index to start with.
    :param increment: Index increment. Use `-1` to iterate backwards.
    """
    def wrap(f):
        def wrapped_f(self, sections):
            i = start
            while i < len(sections) and i > -1:
                new_i = f(self, sections, i)
                i = new_i if new_i else i + increment

            return sections

        wrapped_f.__name__ = f.__name__
        return wrapped_f

    return wrap


def split_section_by_regex(section, regex):
    """ Helper method that splits a section into parts using the `regex` matching against\
        the section code """
    if not section.has_code():
        return [section]

    sections = []
    start = 0
    for match in regex.finditer(section["code_text"]):
        code = section["code_text"][start:match.start()]
        if code.strip():
            sections.append(Section(code_text=code))
        sections.append(Section(docs_text=match.group(1)))
        start = match.end()

    remains = section["code_text"][start:].strip('\n')
    if remains.strip():
        sections.append(Section(code_text=remains))

    return sections


def split_code_by_pos(i, pos, sections):
    code = sections[i]["code_text"]
    section_1 = sections[i].copy()
    section_1['code_text'] = code[:pos]
    section_2 = sections[i].copy()
    section_2['docs_text'] = ""
    section_2['code_text'] = code[pos:]

    sections[i:i+1] = [section_1, section_2]
    return sections



# -*- coding: utf-8 -*-

"""
![Pyccoon](pyccoon.svg)

"**Pyccoon**" is a side-to-side documentation generator.

It descended from [Pycco](https://github.com/fitzgen/pycco) - a Python port of\
[Docco](http://jashkenas.github.com/docco/):\
the original quick-and-dirty, hundred-line-long, literate-programming-style\
documentation generator. It produces HTML that displays your comments\
alongside your code. Comments are passed through \
[Markdown](http://daringfireball.net/projects/markdown/syntax), while code is\
passed through [Pygments](http://pygments.org/) for syntax highlighting.

== This website is the result of running Pyccoon against its source. ==

Pyccoon generates the documentation folder structured correspondingly to the code. To create \
documentation `docs` for the project in `src` folder, run the following:

    pyccoon -s src -d docs

[Pyccoon](https://github.com/ckald/pyccoon) is released on GitHub under the MIT license.
"""

__author__ = 'Andrii Magalich'
__email__ = 'andrew.magalich@gmail.com'
__version__ = '0.1.2'

from .pyccoon import Pyccoon



import os
import time


class cached_property(object):
    """
    Descriptor (non-data) for building an attribute on-demand on first use.
    """
    def __init__(self, factory):
        """
        <factory> is called such: factory(instance) to build the attribute.
        """
        self._attr_name = factory.__name__
        self._factory = factory

    def __get__(self, instance, owner):
        # Build the attribute.
        attr = self._factory(instance)

        # Cache the value; hide ourselves.
        setattr(instance, self._attr_name, attr)

        return attr


def shift(list, default):
    """
    Shift items off the front of the `list` until it is empty, then return
    `default`.
    """

    try:
        return list.pop(0)
    except IndexError:
        return default


def ensure_directory(directory):
    """ === Ensure directory ===
        Ensure that the destination directory exists."""

    if not os.path.isdir(directory):
        os.makedirs(directory)


def monitor(path, func):
    """Monitor each source file and re-generate documentation on change."""

    # The watchdog modules are imported in `main()` but we need to re-import\
    # here to bring them into the local namespace.
    import watchdog.events
    import watchdog.observers

    class RegenerateHandler(watchdog.events.FileSystemEventHandler):
        """A handler for recompiling files which triggered watchdog events"""
        def on_any_event(self, event):
            """Regenerate documentation for a file which triggered an event"""
            # Re-generate documentation from a source file if it was listed on\
            # the command line. Watchdog monitors whole directories, so other\
            # files may cause notifications as well.
            func()

    # Set up an observer which monitors all directories for files given on\
    # the command line and notifies the handler defined above.
    event_handler = RegenerateHandler()
    observer = watchdog.observers.Observer()
    observer.schedule(event_handler, path=path, recursive=True)

    # Run the file change monitoring loop until the user hits Ctrl-C.
    observer.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()



# -*- coding: utf-8 -*-

import re
import os

import pygments
from pygments import lexers, formatters

from markdown import markdown
from .. import markdown_extensions

from ..utils import cached_property
from .utils import Section, ParsingStrategy, iterate_sections,\
    split_section_by_regex, split_code_by_pos


class Language(object):
    """
    == Pyccoon Language definition ==

    This class governs all source file parsing routines. Due to differences in programming \
    languages, an extensible parsing `strategy` is required \
    (see [[./utils.py#parsing-strategy]])
    """

    extensions = []
    scope_keywords = []
    filename_substitutes = {}
    markdown_extensions = [
        markdown_extensions.LinesConnector(),
        markdown_extensions.SaneDefList(),
        markdown_extensions.Todo(),
        markdown_extensions.Pydoc(),
        markdown_extensions.AutoLinkExtension(),
        "def_list", "fenced_code", 'codehilite'
    ]

    @property
    def name(self):
        return self.__class__.__name__

    @cached_property
    def lexer(self):
        """ Pygments lexer corresponding to the language """
        return lexers.get_lexer_by_name(self.name.lower())

    def highlight(self, code, formatter="html"):
        """ Use pygments to highlight the `code` """
        return pygments.highlight(
            code, self.lexer,
            formatters.get_formatter_by_name(formatter)
        )

    def markdown(self, docs):
        return markdown(docs, extensions=self.markdown_extensions)

    def transform_filename(self, filename):
        """
        Filename transformation according to language specifics. If `filename_substitutes` are \
        defined, the filename can be replaced accordingly. For example, `Python` module's\
        `__init__.py` corresponds to the index file of the folder and should be turned into\
        `index.html`

        If no `filename_substitutes` declared, the filename extension will be replaced by `.html`.
        """
        if filename in self.filename_substitutes:
            return self.filename_substitutes[filename]

        for extension in self.extensions:
            if filename.endswith(extension):
                return filename + ".html"

        return filename

    def strategy(self):
        """ Language parsing strategy - i.e., a list of methods to be applied to the code \
            to derive a properly formatter set of docs-code sections """
        return ParsingStrategy(self.set_sections_levels, self.strip_docs_indentation,
                               self.set_sections_levels, self.merge_up,
                               self.set_sections_levels, self.merge_down,
                               self.set_sections_levels, self.absorb)

    def parse(self, code, add_lineno=True):
        """ Apply `self.strategy()` to the `code` """
        sections = [Section(code_text=code)]

        for method in self.strategy():
            sections = method(sections)

        # Strip empty sections
        sections = [section for section in sections if section.has_code() or section.has_docs()]

        return sections

    @iterate_sections(start=0)
    def debug_docs(self, sections, i):
        print(sections[i]['docs_text'])

    @iterate_sections(start=0)
    def debug_code(self, sections, i):
        print(sections[i]['code_text'])

    @iterate_sections(start=0)
    def set_sections_levels(self, sections, i):
        if sections[i]["code_text"]:
            indent = re.match(r"^([ \t]*)", sections[i]["code_text"]).group(1)
            sections[i]["level"] = len(indent)
        elif i > 0:
            sections[i]["level"] = sections[i-1]['level']

    @iterate_sections(start=0)
    def strip_docs_indentation(self, sections, i):
        indent = re.match(r"^([ \t]*)", sections[i]["docs_text"], re.M).group(1)

        sections[i]["docs_text"] = re.compile(r"^{0}".format(indent), re.M)\
            .sub("", sections[i]["docs_text"])

    @iterate_sections()
    def merge_up(self, sections, i):
        """ Suck up the documentation added right under the scope-defining lines (e.g., class or \
            function definition) """
        if not sections[i].has_code() and not sections[i-1].has_docs() and sections[i-1].has_code():

            prev_line = sections[i-1]["code_text"].strip().split("\n")[-1].strip()
            # If previous line of code contains one of the `scope_keywords` - merge last 2 sections
            if any([re.match(x, prev_line) for x in self.scope_keywords]):
                sections[i-1]["docs_text"] = sections[i]["docs_text"]
                sections[i:i + 1] = []

    @iterate_sections()
    def merge_down(self, sections, i):
        """ Merge the documentation placed above the code (just like the next comment) """

        # if there was no code, but were docs - merge
        if not sections[i-1].has_code() and sections[i-1].has_docs()\
                and not sections[i].has_docs() and sections[i].has_code():
            sections[i-1]["code_text"] = sections[i]["code_text"]
            sections[i-1]["scope"] = sections[i]["scope"] or sections[i-1]["scope"]
            sections[i:i+1] = []

    @iterate_sections()
    def absorb(self, sections, i):
        """ Absorb next code-only section if it lies deeper than the current one (that has docs)"""
        if not sections[i].has_docs() and sections[i]['level'] > sections[i-1]['level']:
            sections[i-1]['code_text'] = sections[i-1]['code_text'].rstrip('\n') \
                + '\n\n' + sections[i]['code_text'].lstrip('\n')
            sections[i:i+1] = []


class InlineCommentLanguage(Language):
    """
    == Inline commenting mixins ==

    Language mixin for separate inline comments and whole stacks of them.
    """

    inline_delimiter = "#"

    def strategy(self):
        base_strategy = super(InlineCommentLanguage, self).strategy()
        base_strategy.insert(0, self.parse_inline)
        return base_strategy

    @cached_property
    def inline_re(self):
        """
            ^\s*{0}\s*(.+$)
            (^[ \t]*{0}(.*)$)+
        """
        return re.compile(r"((^[ \t]*{0}.*\n)+)".format(self.inline_delimiter), flags=re.M)

    @property
    def divider_text(self):
        # The dividing token we feed into Pygments, to delimit the boundaries between sections.
        return "\n{0}DIVIDER\n".format(self.inline_delimiter)

    @property
    def divider_html(self):
        # The mirror of `divider_text` that we expect Pygments to return. We can split \
        # on this to recover the original sections.
        return re.compile(r'\n*<span class="c[1]?">{0}DIVIDER</span>\n*'
                          .format(self.inline_delimiter))

    @iterate_sections(start=0)
    def parse_inline(self, sections, i):
        new_sections = split_section_by_regex(sections[i], self.inline_re)
        for j, section in enumerate(new_sections):
            new_sections[j]["docs_text"] = re.compile(
                r"^[ \t]*{0}".format(self.inline_delimiter), re.M
            ).sub("", new_sections[j]["docs_text"])

        sections[i:i+1] = new_sections


class MultilineCommentLanguage(Language):
    """
    == Multiline commenting mixins ==

    Language mixin for multiline comments. Some languages also have another syntax entity\
    called "docblocks" - they probably should be treated separately, although they are usually\
    captured along with multiline comments.
    """

    multistart = '"""'
    multiend = '"""'

    @property
    def multiline_re(self):
        return re.compile(r'^(\s*{start}((?!{end})[\s\S])*){end}'
                          .format(start=self.multistart, end=self.multiend), flags=re.M)

    @property
    def multiline_delimiters(self):
        return [self.multistart, self.multiend]

    def strategy(self):
        base_strategy = super(MultilineCommentLanguage, self).strategy()
        base_strategy.insert(0, self.parse_multiline)
        return base_strategy

    @iterate_sections(start=0)
    def parse_multiline(self, sections, i):
        sections[i:i+1] = split_section_by_regex(sections[i], self.multiline_re)
        sections[i]["docs_text"] = re.sub(r"^\n*(\s*){0}".format(self.multistart),
                                          r"\1",
                                          sections[i]["docs_text"])


class IndentBasedLanguage(Language):

    """
    == Mixins for indent-based languages (Python, Ruby, etc.) ==

    In indent-based languages it is quite easy to find a proper place to split the code section: \
    basically, whenever an indent of the line becomes smaller, than the indent of the first line \
    of the section - split up.

    TODO: Consider using some preprocessor instead of literal matching of the indentation. For \
        example, https://github.com/sirthias/parboiled/wiki/Indentation-Based-Grammars, \
        https://github.com/Cirru/cirru-parser
    """

    @iterate_sections(start=0)
    def split_by_scopes(self, sections, i):
        indent = re.match(r"^(\s*)", sections[i]["code_text"].strip("\n")).group(1)

        regex = re.compile(r"^(\s{{0,{0}}}\S)".format(len(indent) - 1), flags=re.M)
        match = regex.search(sections[i]["code_text"], pos=len(indent) + 1)

        if match:
            sections = split_code_by_pos(i, match.start(), sections)
            sections[i]['level'] = len(indent)
            sections[i+1]['level'] = len(match.group(1).strip("\n"))

        regex = re.compile(r"({0})".format("|".join(self.scope_keywords)), flags=re.M)
        match = regex.search(sections[i]["code_text"])

        if match and match.start() == 0:
            sections[i]['scope'] = match.group(1).strip()
            match = regex.search(sections[i]["code_text"], pos=match.start() + 1)

        if match:
            sections = split_code_by_pos(i, match.start(), sections)
            sections[i+1]['code_text'] = sections[i+1]['code_text'].strip('\n')
            sections[i+1]['scope'] = match.group(1).strip()

    def strategy(self):
        base_strategy = super(IndentBasedLanguage, self).strategy()
        base_strategy.insert_before('merge_up', self.split_by_scopes)
        return base_strategy


# == Mixins for brace-based languages (C/C++, JavaScript, PHP, etc.) ==

class BraceBasedLanguage(Language):

    @iterate_sections(start=0)
    def split_by_scopes(self, sections, i):
        """ Split the code sections by `scope_keywords` of the language
            TODO: consider splitting also by braces interiors"""

        regex = re.compile(r"^({0})".format("|".join(self.scope_keywords)), flags=re.M)
        match = regex.search(sections[i]["code_text"])

        if match and match.start() == 0:
            match = regex.search(sections[i]["code_text"], pos=match.start() + 1)

        if match:
            split_code_by_pos(i, match.start(), sections)
            sections[i+1]['code_text'] = sections[i+1]['code_text'].strip('\n')
            sections[i+1]['scope'] = match.group(2)

    def strategy(self):
        base_strategy = super(BraceBasedLanguage, self).strategy()
        base_strategy.insert_before('merge_up', self.split_by_scopes)
        return base_strategy


# == Specific languages definitions ==

class C(BraceBasedLanguage, InlineCommentLanguage, MultilineCommentLanguage):
    """
    === C/C++ ===

    Styling of the C/C++ code is largely historical and oriented on reading hopelessly long codes\
    on the old terminal screens.

    TODO: detect whole style docs blocks, for example, boxes:

    ```c
        ////////////////////////////
        ////// Nice comment! ///////
        ////////////////////////////

        /***************************
        **** We love ASCII-art! ****
        ***************************/
    ```

    """
    extensions = [".c", ".cpp", ".h", ".cc"]
    inline_delimiter = "//"
    multistart = r"/\*+"
    multiend = r"\*+/"

    scope_keywords = [r"\s*({0}) \w+".format(word) for word in
                      ["class", "function", "namespace",
                       "public", "private", "protected", "abstract", "inline", "virtual",
                       "void", "int", "double", "bool",  "float"]]

    # FIXME: this redefinition brakes the whole extension
    # ```python
    # def __init__(self, *args, **kwargs):
    #     super(C, self).__init__(*args, **kwargs)
    #     for i, ext in enumerate(self.markdown_extensions):
    #         if isinstance(ext, markdown_extensions.LineConnector):
    #             self.markdown_extensions[i] = \
    #                 markdown_extensions.LineConnector(regex=r"([\w\.])[ \t]*\n[ \t]*(\w)")
    # ```

    @iterate_sections(start=0)
    def strip_commenting_design(self, sections, i):
        sections[i]["docs_text"] = re.compile(r"^[ \t]*(\/+|\*+)(.*)$", re.M)\
            .sub(r"\2", sections[i]["docs_text"])

    def strategy(self):
        base_strategy = super(C, self).strategy()
        base_strategy.insert_before('strip_docs_indentation', self.strip_commenting_design)
        base_strategy.insert_after('strip_docs_indentation', self.split_by_scopes)
        return base_strategy


class JavaScript(C):
    """
    === JavaScript ===
    JavaScript is largely identical to C/C++, although it has far less scope keywords and far more\
    flexibility in defining functions and objects.

    TODO: test function-defining scopes

    ```javascript
        function name(args) { ... }
        name = function(args) { ... }
    ```

    """

    extensions = [".js"]

    scope_keywords = [r"\s*(class)", r"^.*(function)[ \t]*\("]


class PHP(C):
    extensions = [".php"]

    def strategy(self):
        base_strategy = super(PHP, self).strategy()
        base_strategy.delete('merge_up')
        return base_strategy


class Python(IndentBasedLanguage, MultilineCommentLanguage, InlineCommentLanguage):
    """
    === Python ===
    Obviously, Python language parsing is a best-developed part of Pyccoon.

    TODO: support also `'''` comment delimiters.
    """
    extensions = [".py", ".pyx"]
    inline_delimiter = "#"
    multistart = '"""'
    multiend = '"""'

    # `__init__.py` files can perfectly serve as modules index files.
    filename_substitutes = {
        "__init__.py": "index.html"
    }

    scope_keywords = [r"^\s*(def) ", r"^\s*(class) ", r"^\s*(@)\w+"]

    def strategy(self):
        base_strategy = super(Python, self).strategy()
        base_strategy.insert_before('absorb', self.python_absorb)
        return base_strategy

    @iterate_sections()
    def python_absorb(self, sections, i):
        """
        Python decorators are the tricky part of proper parsing of the source file into \
        sections of docs and code.

        Whenever a decorator section occurs, it should be merged not into the previous sections,
        but into the next.
        """

        if '@' in sections[i-1]['scope']:
            sections[i]['docs_text'] = sections[i-1]['docs_text'] + sections[i]['docs_text']
            sections[i]['code_text'] = sections[i-1]['code_text'] + sections[i]['code_text']
            sections[i-1:i+1] = [sections[i]]
            return i


class Ruby(IndentBasedLanguage, InlineCommentLanguage, MultilineCommentLanguage):
    """
    === Ruby ===
    Mostly identical to Python.

    TODO: Actually, Ruby is crazy and supports unbelievable variety of multiline comment syntaxes:

    ```python
        multistart = ["=begin", "<<-DOC", "\"", "__END__"]
        multiend = ["=end", "DOC", "\"", ""]
    ```
    will have to rethink multiline comments capturing to support them all
    """
    extensions = [".rb"]
    inline_delimiter = "#"
    multistart = "=begin"
    multiend = "=end"

    scope_keywords = [r"^\s*(module) ", r"^\s*(class) ", r"^\s*(def) "]

"""
== Languages in development ==

```python
class CoffeScript(InlineCommentLanguage, MultilineCommentLanguage):
    extensions = [".coffee"]
    name = "Coffee-Script"
    inline_delimiter = "#"
    multistart = "###"
    multiend = "###"


class Perl(InlineCommentLanguage):
    extensions = [".pl"]
    inline_delimiter = "#"


class SQL(InlineCommentLanguage):
    extensions = [".sql"]
    inline_delimiter = "--"


class Scheme(InlineCommentLanguage, MultilineCommentLanguage):
    extensions = [".scm"]
    inline_delimiter = ";;"
    multistart = "#|"
    multiend = "|#"


class Lua(InlineCommentLanguage, MultilineCommentLanguage):
    extensions = [".lua"]
    inline_delimiter = "--"
    multistart = "--[["
    multiend = "--]]"


class Erlang(InlineCommentLanguage):
    extensions = [".erl"]
    inline_delimiter = "%%"


class Tcl(InlineCommentLanguage):
    extensions = [".tcl"]
    inline_delimiter = "#"


class Haskell(InlineCommentLanguage, MultilineCommentLanguage):
    extensions = [".hs"]
    inline_delimiter = "--"
    multistart = "{-"
    multiend = "-}"

languages = [CoffeScript, Perl, SQL, C, PHP,  JavaScript, Ruby, Python, Scheme,
             Lua, Erlang, Tcl, Haskell]
```
"""

extensions_mapping = {}

languages = [Python, PHP, C, JavaScript, Ruby]

for language in languages:
    instance = language()
    for extension in instance.extensions:
        extensions_mapping[extension] = instance


def get_language(source, code, language=None):
    """Get the current language we're documenting, based on the extension."""

    if language is not None:
        for l in extensions_mapping.values():
            if l.name == language:
                return l
        else:
            raise ValueError("Unknown forced language: " + language)

    m = re.match(r'.*(\..+)', os.path.basename(source))

    if m and m.group(1) in extensions_mapping:
        return extensions_mapping[m.group(1)]
    else:
        lang = lexers.guess_lexer(code).name.lower()
        for l in extensions_mapping.values():
            if l.name == lang:
                return l
        else:
            return None
